//calibrate.cc (C) Copyright 2007, Phil Underwood

// This file is part of PonyTrainer.

// PonyTrainer is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// PonyTraine is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with PonyTrainer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

/* calibrate.cc */
// this file is to read in data and perform the calibration step
// hopefully this should be a lot faster than the python version


#include <deque>
#include <math.h>
#include <iostream>
#include <numeric>
#include <vector>
#include <string.h>
#define LOO
#define INVAR_WEIGHT 50
#define ASSESS_MATRIX AssessMatrixOld

double bestM[12] = {1,0,0,
		   0,1,0,
		   0,0,1,
		   0,0,0}; // calibration matrix for Mag sensors
double bestG[12] = {1,0,0,
		   0,1,0,
		   0,0,1,
		   0,0,0}; // calibration matrix for Mag sensors
double identity[12] = {1,0,0,
		   0,1,0,
		   0,0,1,
		   0,0,0}; // calibration matrix for Mag sensors

typedef std::vector<double> doublevec;

class dataline {
public:
  char name[10];
  double Md[3];
  double Gd[3];
  double V[3];
  double bearing, inclination;
  dataline();
  dataline(char* line);
};

double average(doublevec d) {
  return (std::accumulate(d.begin(),d.end(),double(0))/d.size());
}

double stdev(doublevec d) {
  double sum = std::accumulate(d.begin(),d.end(),double(0));
  double sum_sq = std::inner_product(d.begin(),d.end(),d.begin(),double(0));
  double N = d.size();
  return sqrt(((N * sum_sq) - (sum*sum))/(N * (N-1.0)));
}

dataline::dataline() {
}

dataline::dataline(char* line) {
  if (sscanf(line,"%s %lf %lf %lf %lf %lf %lf %lf %lf",
	     name,
	     &(Gd[0]),   
	     &(Gd[1]),
	     &(Gd[2]),
	     &(Md[0]),
	     &(Md[1]),
	     &(Md[2]), 
	     &bearing,
	     &inclination
	     ) < 8) std::cerr << "Bad reading!";
  double b = bearing * M_PI / 180;
  double i = inclination * M_PI / 180;

  V[0] = cos(b) * cos(i);
  V[1] = sin(b) * cos(i);
  V[2] = sin(i);
}

void Normalise(double *x) {
	double _t;
	_t = sqrt(x[0] * x[0] + x[1] * x[1] + x[2] * x[2]); 
	x[0] /= _t; 
	x[1] /= _t; 
	x[2] /= _t;
}

// a = b * c; b is 3v, c is 34m
void Multiply(double *a, double *b, double *c) {
	a[0] = b[0] * c[0] + b[1] * c[3] + b[2] * c[6] + c[9]; 
	a[1] = b[0] * c[1] + b[1] * c[4] + b[2] * c[7] + c[10];
	a[2] = b[0] * c[2] + b[1] * c[5] + b[2] * c[8] + c[11];
} 

// a = b x c
void CrossProduct(double *a, double *b, double *c) {
	a[0] = b[1] * c[2] - b[2] * c[1]; 
	a[1] = b[2] * c[0] - b[0] * c[2]; 
	a[2] = b[0] * c[1] - b[1] * c[0];
}

double Magnitude (double *a) {
  return sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2]);
}

void 
GetReadings(double *M, double *G, double *Mc, double *Gc, double *V, double *AM = 0, double *AG = 0, double *dip = 0) {
	// Calculate what our bearing and inclination should be...
	double fM[3];
	double down[3];
	double east[3];
	double north[3];
	Multiply(down,G,Gc);
	if (AG) *AG = Magnitude(down);
	Normalise(down);
	down[0] *= -1; // I still don't know why I need to do this!
	down[1] *= -1;
	down[2] *= -1;
	Multiply(fM,M,Mc);
	CrossProduct(east,down,fM);
	if (AM) *AM = Magnitude(fM);
	if (dip) *dip = Magnitude(east);
	Normalise(east);
	CrossProduct(north,east,down);
      	Normalise(north);
	V[0] = north[0];
	V[1] = east[0];
	V[2] = down[0];
}


void GetBearings(double *M, double *G, double *Mc, double *Gc, double *compass, double *clino) {
  double V[3];
  GetReadings(M,G,Mc,Gc,V);
  *compass = atan2(V[1],V[0]) * (180 / M_PI);
  *clino = atan2(V[2],sqrt(V[0] * V[0] + V[1] * V[1])) * (180 / M_PI);
  if (*compass < 0) *compass += 360;
}

double
GetError(double *V1, double *V2) {
  double temp;
//    temp = fabs(V1[0] - V2[0]);
//    temp +=fabs(V1[1] - V2[1]);
//    temp +=fabs(V1[2] - V2[2]);
  temp = (V1[0] - V2[0]) * (V1[0] - V2[0]);
  temp +=(V1[1] - V2[1]) * (V1[1] - V2[1]);
  temp +=(V1[2] - V2[2]) * (V1[2] - V2[2]);
  return temp;
}

double
GetDegError(dataline *d,double *Mc, double *Gc) {
  double V[3];
  GetReadings(d->Md,d->Gd,Mc,Gc,V);
  return sqrt(GetError(d->V,V)) * 180.0/M_PI;
}

double
AssessMatrixOld(double *Mc, double *Gc, std::deque<dataline> data) {
  double V[3];
  double result = 0;
  for (std::deque<dataline>::iterator d = data.begin(); d < data.end(); ++d) {
    GetReadings(d->Md,d->Gd,Mc,Gc,V);
    result += GetError(d->V,V);
  }
  return result;
}

double
AssessMatrixNew(double *Mc, double *Gc, std::deque<dataline> data) {
  double V[3];
  double dip, AM, AG;
  double SX[3] = {0,0,0};
  double SX2[3] = {0,0,0};
  double MSX,MSX2,GSX,GSX2, DSX, DSX2;
  double result = 0;
  int counter = 0;
  int j = 0;
  int k = 0;
  double oldclino = -100;
  double oldbearing = -1;
  MSX = MSX2 = GSX = GSX2 = DSX = DSX2 = 0;
  for (std::deque<dataline>::iterator d = data.begin(); d < data.end(); ++d) {
    GetReadings(d->Md,d->Gd,Mc,Gc,V,&AM, &AG, &dip);
    if ((oldclino != d->inclination) || (oldbearing != d->bearing)) {
      oldclino = d->inclination;
      oldbearing = d->bearing;
      if (counter > 0) {
	for (int i =0; i< 3; ++i){
	  double mean = SX[i]/counter;
	  if ((k % 4)==0) result += SX2[i] - (counter*mean*mean);	
	  SX[i] = SX2[i] = 0;
	}
	k++;
	counter = 0;
      }
    }
    GSX += AG;
    GSX2 += AG * AG;
    MSX += AM;
    MSX2 += AM * AM;
    DSX += dip;
    DSX2 += dip * dip;
    for (int i = 0; i < 3 ; ++i) {
      SX[i] += V[i];
      SX2[i] += V[i] * V[i];
    }
    ++counter;
    ++j;
  }
  for (int i =0; i< 3; ++i){
    double mean = SX[i]/counter;
    result += SX2[i] - (counter*mean*mean);
  }
  result += (MSX2 - MSX*MSX/j) * INVAR_WEIGHT;
  result += (GSX2 - GSX*GSX/j) * INVAR_WEIGHT;
  result += (DSX2 - DSX*DSX/j) * INVAR_WEIGHT;
  return result;
}


void Teeter(double *Mc, double *Gc, int index, double delta, int axis) {
  switch(axis) {
  case 0: 
    Mc[index] = Mc[index] + delta;
    break;
  case 1:
    Mc[index] = Mc[index] - delta;
    break;
  case 2:
    Gc[index] = Gc[index] + delta;
    break;
  case 3:
    Gc[index] = Gc[index] - delta;
    break;
  }
}

double degError(double val, int count) {
  return sqrt(val/count)*180/M_PI;
}

double Calibrate(std::deque<dataline> data, 
	       int iterations, 
	       int limited, 
	       double delta) {
  double bestR;
  double oldR;
  double result;
  long int tally = 0;
  bestR = ASSESS_MATRIX(bestM,bestG,data);
  oldR = bestR + 1;
  for (int scale = 0; scale < iterations; ++scale) {
    #ifndef LOO
    std::cerr << scale << ":" << bestR << ":" << AssessMatrixOld(bestM,bestG,data) << ":" << tally << "\n";
    #endif
    while (oldR > bestR) {
      oldR = bestR;
      for (int dir = 0; dir < 12; limited ? (dir<8?dir+=4:++dir) : ++dir) {
	for (int x = 0; x < 4; ++x) {
	  double Mc[12];
	  double Gc[12];
	  memcpy(Mc,bestM,sizeof(double) * 12);
	  memcpy(Gc,bestG,sizeof(double) * 12);
	  Teeter(Mc,Gc,dir,delta,x);
	  result = ASSESS_MATRIX(Mc,Gc,data);
	  tally++;
	  if (result < bestR) {
	    memcpy(bestM,Mc,sizeof(double) * 12);
	    memcpy(bestG,Gc,sizeof(double) * 12);
	    bestR = result;
	    //	    std::cerr << bestR << "\n";
	  }
	   
	}
      }
    }
    delta = delta/2;
    oldR = bestR + 1;
  }
  std::cerr << tally << "\n";
  return degError(bestR,data.size());
}

uint32_t f2l(float f) {
  uint32_t temp;
  temp = *((unsigned int*)&f);
  return temp;
}

int
main(int ARGC, char **ARGV) {
  std::deque<dataline> data;
  char text[256];
  char loo = 0;
  if (ARGC>1) {
    if (strcmp(ARGV[1],"-l")==0) {
      loo = 1;
    }
  }
  while (std::cin) {
    std::cin.getline(text,256);
    if (strlen(text) > 6) data.push_back(dataline(text));
  }
  doublevec orth_err,comp_err,clino_err;
  //data.pop_back();
  // leave one out analysis
  if (loo) {
#define STEP 4
    for (int x = data.size()/STEP; x > 0; --x) {
      memcpy(bestM,identity,sizeof(double)*12);
      memcpy(bestG,identity,sizeof(double)*12);
      dataline temp[STEP];
      for (int y = 0; y < STEP; ++y) {
	temp[y] = data.back();
	data.pop_back();
      }
      Calibrate(data,12,1,1.0);
      double comp,clino;
      double g = Calibrate(data,18,0,0.1);
      for (int y =0 ; y < STEP; ++y) {
	GetBearings(temp[y].Md,temp[y].Gd,bestM,bestG,&comp,&clino);
	double f = GetDegError(&temp[y],bestM,bestG);
	orth_err.push_back(f);
	comp_err.push_back(temp[y].bearing - comp);
	clino_err.push_back(temp[y].inclination - clino);
	fprintf(stdout,"%s\t%5.1f\t%5.1f\t%+04.1f\t%+04.1f\t%5.2f\n",temp[y].name,temp[y].bearing,comp,temp[y].inclination,clino,f);
	fflush(stdout);
	data.push_front(temp[y]);
      }
    }
    printf("# %g\n",average(orth_err));
    fprintf(stderr,"Average: %g\nStDev: %g\n",average(orth_err),stdev(orth_err));
    fprintf(stderr,"Compass Average: %g\nCompass StDev: %g\n",average(comp_err),stdev(comp_err));
    fprintf(stderr,"Clino Average: %g\nClino StDev: %g\n",average(clino_err),stdev(clino_err));
    orth_err.clear();
    comp_err.clear();
    clino_err.clear();
  } else {
    memcpy(bestM,identity,sizeof(double)*12);
    memcpy(bestG,identity,sizeof(double)*12);
    Calibrate(data,12,0,1.0);  
    double g = Calibrate(data,18,0,0.1);
    for (int x  = 0; x < data.size(); ++x) {
      dataline temp = data[x];
      double comp,clino;
      GetBearings(temp.Md,temp.Gd,bestM,bestG,&comp,&clino);
      double f = GetDegError(&temp,bestM,bestG);
      orth_err.push_back(f);
      comp_err.push_back(temp.bearing - comp);
      clino_err.push_back(temp.inclination - clino);
      fprintf(stderr,"%5.2f\t(%5.1f:%5.1f)\t(%+04.1f:%+04.1f)\n",f,temp.bearing,comp,temp.inclination,clino);
    }
    fprintf(stderr,"Average: %g\nStDev: %g\n",average(orth_err),stdev(orth_err));
    fprintf(stderr,"Compass Average: %g\nCompass StDev: %g\n",average(comp_err),stdev(comp_err));
    fprintf(stderr,"Clino Average: %g\nClino StDev: %g\n",average(clino_err),stdev(clino_err));
    for (int i = 0; i < 12; ++i) {
      printf("SM%1X %08X\n",i,f2l(float(bestM[i])));
    }
    std::cerr << "\n";
    for (int i = 0; i < 12; ++i) {
      printf("SG%1X %08X\n",i,f2l(float(bestG[i])));
    }
    printf("# %g\n",average(orth_err));
  }
  return 0;
}
